''' ORM and ODM tool for FastApi simplification. It enables the user to define only 1 PyDantic models and work with
 data on multiple sources '''
import inspect
from enum import Enum
from typing import List, Optional, TypeVar, Union, Generic
from odim.helper import awaited

from pydantic import BaseModel, Field, root_validator
from pydantic.generics import GenericModel
from datetime import datetime
from odim.helper import get_config, get_connection_info, get_connector_for_model

all_json_encoders = {
  datetime: lambda dt: dt.isoformat()
}

T = TypeVar('T')

class SearchParams(BaseModel):
  ''' Describes how to search the details '''
  offset : int = 0
  limit : int = 25
  sort : Optional[str] = Field(default=None, description="Order by field list, separated by comma with - signifying descending order. e.g. name,-created_at  will order by name ASC and created_at DESC", regex="[,a-zA-Z0-9_-]*")


class SearchResponse(GenericModel, Generic[T]):
  search : dict = Field(description="The search data that was performed")
  total : int =  Field(description="The total number of results")
  results : List[T]

  class Config:
    json_encoders = all_json_encoders


class OkResponse(BaseModel):
  ok : bool = Field(default=True)


class Operation(Enum):
  exact = "__is"
  isnot = "__not"
  contains = "__contains"
  gt = "__gt"
  gte = "__gte"
  lt = "__lt"
  lte = "__lte"
  null = "__null"

def parse_fieldop(field):
  for o in Operation:
    if field.endswith(o.value):
      return field.replace(o.value, ""), o
  return field, Operation.exact


class BaseOdimModel(BaseModel):

  @root_validator(pre=True)
  def generic_validators_pre(cls, values):
    if hasattr(cls, "Config") and hasattr(cls.Config, "odim_hooks"):
      for fnc in cls.Config.odim_hooks.get("pre_validate",[]):
        values = awaited(fnc(values))
    return values

  @root_validator()
  def generic_validators_post(cls, values):
    if hasattr(cls, "Config") and hasattr(cls.Config, "odim_hooks"):
      for fnc in cls.Config.odim_hooks.get("post_validate",[]):
        values = awaited(fnc(values))
    return values


  async def save(self, *args, **kwargs):
    return await Odim(self).save(*args, **kwargs)

  async def update(self, *args, **kwargs):
    return await Odim(self).update(*args, **kwargs)

  async def delete(self, force_harddelete = False):
    return await Odim(self).delete(force_harddelete = force_harddelete)

  @classmethod
  async def find(cls, *args, **kwargs):
    return await Odim(cls).find(*args, **kwargs)

  @classmethod
  async def count(cls, *args, **kwargs):
    return await Odim(cls).count(*args, **kwargs)

  @classmethod
  async def get(cls, *args, **kwargs):
    return await Odim(cls).get(*args, **kwargs)


class Odim(object):
  ''' Initiates the wrapper to communicate with backends based on the pydantic model Config metaclass '''
  protocols = []
  model = None
  instance = None

  def __new__(cls, model):
    odimclass = get_connector_for_model(model)
    return super(Odim, cls).__new__(odimclass)


  def __init__(self, model : Union[BaseModel, BaseModel.__class__]):
    if inspect.isclass(model):
      self.model = model
    else:
      self.model = model.__class__
      self.instance = model


  def get_connection_identifier(self):
    if hasattr(self.model, 'Config'):
      if hasattr(self.model.Config, 'db_name'):
        return self.model.Config.db_name
      if hasattr(self.model.Config, 'db_uri'):
        return self.model.Config.db_uri
    for key in get_config('DATABASES', default={}).keys():
      cp = get_connection_info(key)
      if cp.protocol in self.protocols:
        return key
    raise AttributeError("missing database definition")

  def softdelete(self):
    if hasattr(self.model, 'Config') and hasattr(self.model.Config, 'softdelete'):
      return getattr(self.model.Config,'softdelete')

  def execute_hooks(self, hook_type, obj, *args, **kwargs):
    if hasattr(self.model, "Config") and hasattr(self.model.Config, "odim_hooks"):
      for fnc in self.model.Config.odim_hooks.get(hook_type,[]):
        obj = awaited(fnc(obj, *args, **kwargs))
    return obj


  def has_hooks(self, *hook_types):
    ''' Whether there are hooks '''
    for hk in hook_types:
      if hasattr(self.model, "Config") and hasattr(self.model.Config, "odim_hooks") and len(self.model.Config.odim_hooks.get(hk, []))>0:
        return True
    return False


  async def save(self, extend_query : dict= {}, include_deleted : bool = False):
    ''' Saves the document and returns its identifier '''
    raise NotImplementedError("Method not implemented for this connector")


  async def update(self, extend_query : dict= {}, include_deleted : bool = False):
    ''' Saves only the changed fields leaving other fields alone '''
    raise NotImplementedError("Method not implemented for this connector")


  async def get(self, id : str, extend_query : dict= {}, include_deleted : bool = False):
    '''
    Retrieves the document by its id
    :param id: id of the docuemnt
    :param extend_query additional search limiters:
    :return: the document as pydantic instance
    '''
    raise NotImplementedError("Method not implemented for this connector")


  def parse_query_operations(self, query : dict):
    ''' Gets the normalized search operations from the query fields '''
    rsp = {}
    for k, v in query.items():
      key, op = parse_fieldop(k)
      rsp[key] = (op, v)
    return rsp


  async def find(self, query : dict, params : SearchParams = None, include_deleted : bool = False):
    ''' Performs search using a dictionary qury to find documents on that particular collection/table
    :param query: dictionary of field:value pairs
    :param params: additional search params like ordering and limit offset
    :return: the list of documents as per pydantic type    '''
    raise NotImplementedError("Method not implemented for this connector")


  async def count(self, query : dict, include_deleted : bool = False) -> int:
    ''' Do the search and count the documents

    :param query: dictionary of field:value pairs
    :return: the number of results '''
    raise NotImplementedError("Method not implemented for this connector")


  async def delete(self, obj : str, extend_query = {}, force_harddelete : bool = False):
    ''' Delete the document from storage '''
    raise NotImplementedError("Method not implemented for this connector")



class NotFoundException(Exception):
  pass


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="odim", # Replace with your own username
    version="0.1.12",
    author="Belda",
    author_email="jakub.belescak@centrum.cz",
    description="Simple Python ORM/ODM specifically designed to be used with Pydantic and FastAPI ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/belda/odim",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires = [
        "pydantic",
        "nest_asyncio"
    ]
)


'''
This module holds the ModelFactory generated pydantic instances
'''
from datetime import datetime
from typing import Optional, Union
from pydantic import Field

from odim.mongo import ObjectId

v= None


import logging
import re
from enum import Enum
from typing import Union

import aiomysql.cursors
from pydantic import BaseModel
from pymysql import escape_string
from pymysql.converters import escape_bytes_prefixed, escape_item

from odim import BaseOdimModel, NotFoundException, Odim, Operation, SearchParams, get_connection_info

log = logging.getLogger("uvicorn")
pools = {}


async def connected_pool(db):
  global pools
  if not db in pools:
    cn = get_connection_info(db)
    if not cn.port:
      cn.port = 3306
    pools[db] =  await aiomysql.create_pool(host=cn.host, port=cn.port, user=cn.username, password=cn.password,
                                            db=cn.db, cursorclass=aiomysql.cursors.DictCursor)
  return pools[db]


class Op(Enum):
  execute = 0
  fetchone = 1
  fetchall = 2


async def execute_sql(db, sql, co : Op = Op.execute):
  log.info(sql)
  pool = await connected_pool(db)
  async with pool.acquire() as conn:
    cursor = await conn.cursor()
    executed = await cursor.execute(sql)
    if co == Op.fetchone:
      return await cursor.fetchone()
    elif co == Op.fetchall:
      return await cursor.fetchall()
    else: #execute
      await conn.commit()
      return cursor

    #TODO handle disconnects and ends
    #TODO pymysql.err.OperationalError



class BaseMysqlModel(BaseOdimModel):
  pass


class OdimMysql(Odim):
  protocols = ["mysql"]

  def escape(self, obj):
    """ Escape whatever value you pass to it"""
    if isinstance(obj, str):
      return "'" + escape_string(obj) + "'"
    if isinstance(obj, bytes):
      return escape_bytes_prefixed(obj)
    return escape_item(obj, getattr(self.model.Config, 'charset', 'utf-8'))


  def get_table_name(self):
    ci = self.get_connection_identifier()
    if hasattr(self.model, 'Config'):
      if hasattr(self.model.Config, 'table_name'):
        cn = self.model.Config.table_name
        return ci, cn
    return ci, self.model.__class__.__name__


  async def get(self, id : str, extend_query : dict= {}, include_deleted : bool = False):
    '''
    Retrieves the document by its id
    :param id: id of the docuemnt
    :param kwargs:
    :return: the document as pydantic instance '''
    #TODO just the desired fields
    db, table = self.get_table_name()
    query = {"id" : self.escape(id), **extend_query}
    if self.softdelete() and not include_deleted:
      query[self.softdelete()] = False
    wh = self.get_where(query)
    rsp = await execute_sql(db, "SELECT * FROM %s WHERE %s" % (escape_string(table), wh), Op.fetchone)
    if not rsp:
      raise NotFoundException()
    ret = self.execute_hooks("pre_init", rsp)
    x = self.model(**ret)
    return self.execute_hooks("post_init", x)

  def get_field_pairs(self, field_dict):
    inss = []
    for k, v in field_dict.items():
      if not re.match("[a-zA-Z0-9_]+", k):
        raise AttributeError("Writing a non ASCII field name")
      if k!="id":
        inss.append( "`"+k+"`="+str(self.escape(v)) )
    return ",".join(inss)

  async def save(self, extend_query : dict= {}, include_deleted : bool = False):
    ''' Saves the document and returns its identifier '''
    db, table = self.get_table_name()
    iii = self.execute_hooks("pre_save", self.instance)
    do = iii.dict(by_alias=True)

    if self.instance.id in (None, ""):
      if self.softdelete() and self.softdelete() not in do:
        do[self.softdelete()] = False
      upff = self.get_field_pairs({**extend_query, **do})
      rsp = await execute_sql(db, "INSERT INTO %s SET %s" % (escape_string(table), upff), Op.execute)
      self.instance.id = rsp.lastrowid
      do = self.execute_hooks("post_save", do, update=False)
      return rsp.lastrowid
    else:
      softdel = {self.softdelete(): False} if self.softdelete() and not include_deleted else {}
      upff = self.get_field_pairs(do)
      whr = self.get_where({"id" : self.instance.id, **softdel, **extend_query})
      sql = "UPDATE %s SET %s WHERE %s" % (escape_string(table), upff, whr)
      rsp = await execute_sql(db, sql, Op.execute)
      do = self.execute_hooks("post_save", do, update=True)
      return self.instance.id


  async def update(self, extend_query : dict= {}, include_deleted : bool = False):
    ''' Updates just the partial document '''
    db, table = self.get_table_name()
    iii = self.execute_hooks("pre_save", self.instance)
    dd = iii.dict(exclude_unset=True, by_alias=True)
    dd_id = dd["id"]
    softdel = {self.softdelete(): False} if self.softdelete() and not include_deleted else {}
    upff = self.get_field_pairs(dd)
    whr = self.get_where({"id" :dd_id, **softdel, **extend_query})
    sql = "UPDATE %s SET %s WHERE %s" % (escape_string(table), upff, whr)
    rsp = await execute_sql(db, sql, Op.execute)
    do = self.execute_hooks("post_save", dd, update=True)


  def get_where(self, query):
    whr = []
    for k, (op, v) in self.parse_query_operations(query).items():
      if not re.match("[a-zA-Z0-9_]+", k):
        raise AttributeError("Searching on a non ASCII field name")
      if op == Operation.exact:
        whr.append( "`"+k+"`="+str(self.escape(v)) )
      elif op == Operation.isnot:
        whr.append( "`"+k+"`!="+str(self.escape(v)) )
      elif op == Operation.contains:
        whr.append( "`"+k+"` LIKE '%"+str(escape_string(v)+"%'") )
      elif op == Operation.gt:
        whr.append( "`"+k+"` > "+str(self.escape(v)) )
      elif op == Operation.gte:
        whr.append( "`"+k+"` >= "+str(self.escape(v)) )
      elif op == Operation.lt:
        whr.append( "`"+k+"` < "+str(self.escape(v)) )
      elif op == Operation.lte:
        whr.append( "`"+k+"` <= "+str(self.escape(v)) )
      elif op == Operation.null:
        if v:
          whr.append( "`"+k+"` IS NULL" )
        else:
          whr.append( "`"+k+"` IS NOT NULL" )
    return  "1" if len(whr) == 0  else " AND ".join(whr)


  async def find(self, query : dict, params : SearchParams = None, include_deleted : bool = False):
    ''' Performs search using a dictionary qury to find documents on that particular collection/table
    :param query: dictionary of field:value pairs
    :param params: additional search params like ordering and limit offset
    :return: the list of documents as per pydantic type    '''
    db, table = self.get_table_name()
    if self.softdelete() and not include_deleted:
      query = {self.softdelete(): False, **query}
    where = self.get_where(query)
    sql_params = ""
    if params:
      if params.sort not in (None, ''):
        sql_params+= " ORDER BY "
        paramslist = []
        for x in params.sort.split(","):
          paramslist.append( (x[1:]+" DESC ") if x[0] == "-" else (x+" ASC ") )
        sql_params+= ",".join(paramslist)
      if params.limit:
        sql_params+= " LIMIT "+str(params.limit)
      if params.offset:
        sql_params+= " OFFSET "+str(params.offset)
    rsp = await execute_sql(db, "SELECT * FROM %s WHERE %s %s" % (escape_string(table), where, sql_params), Op.fetchall)
    rsplist = []
    for row in rsp:
      x2 = self.execute_hooks("pre_init", row)
      m = self.model( **row )
      rsplist.append( self.execute_hooks("post_init", m) )
    return rsplist


  async def count(self, query : dict, include_deleted : bool = False) -> int:
    ''' Do the search and count the documents
    :param query: dictionary of field:value pairs
    :return: the number of results '''
    db, table = self.get_table_name()
    if self.softdelete() and not include_deleted:
      query = {self.softdelete(): False, **query}
    where = self.get_where(query)
    rsp = await execute_sql(db, "SELECT COUNT(*) as cnt FROM %s WHERE %s" % (escape_string(table), where), Op.fetchone)
    return rsp["cnt"]


  async def delete(self, obj : Union[str, int, BaseModel], extend_query : dict= {}, force_harddelete : bool = False):
    ''' Delete the document from storage '''
    db, table = self.get_table_name()
    id = obj if not isinstance(obj, BaseModel) else obj.id
    if self.softdelete() and not force_harddelete:
      whr = self.get_where({"id" : self.escape(id), **extend_query})
      await execute_sql(db, "UPDATE %s SET `%s`=true WHERE %s" % (escape_string(table), self.softdelete(), whr), Op.execute)
    else:
      whr = self.get_where({"id" : self.escape(id), **extend_query})
      await execute_sql(db, "DELETE FROM %s WHERE %s" % (escape_string(table), whr), Op.execute)
    #TODO detect not found





from abc import abstractmethod


class BaseSignals:

  @classmethod
  @abstractmethod
  def pre_init(cls, doc):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def post_init(cls,model):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def pre_save(cls, model):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def post_save(cls, doc):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def pre_validate(cls, model):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def post_validate(cls,model):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def pre_remove(cls, model):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def post_remove(cls, model):
    raise NotImplementedError()